/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.Thread.UncaughtExceptionHandler;
import java.text.MessageFormat;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.iav.gllernen.gllernen.domain.GanglinienLernenParameterLeser;
import de.bsvrz.iav.gllernen.gllernen.portale.DavGanglinienLernenParameterLeser;
import de.bsvrz.iav.gllib.gllib.domain.Archiv;
import de.bsvrz.iav.gllib.gllib.domain.BetriebsMeldungsSender;
import de.bsvrz.iav.gllib.gllib.domain.EreignisKalender;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.ObjektFactory;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienSpeicherClient;
import de.bsvrz.iav.gllib.gllib.portals.DavArchiv;
import de.bsvrz.iav.gllib.gllib.portals.DavBetriebsmeldungsSender;
import de.bsvrz.iav.gllib.gllib.portals.DavChecker;
import de.bsvrz.iav.gllib.gllib.portals.DavEreignisKalender;
import de.bsvrz.iav.gllib.gllib.portals.DavGanglinienSpeicher;
import de.bsvrz.iav.gllib.gllib.portals.DavObjektFactory;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Enthält die main-Methode der Applikation.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class GlLernen implements StandardApplication, UncaughtExceptionHandler {

    /**
     * Einsprungmethode der Applikation <i>Automatisches Lernen Ganglinien</i>.
     *
     * @param args die Kommandozeilenargumente.
     * @see #parseArguments(ArgumentList)
     */
    public static void main(final String[] args) {
        StandardApplicationRunner.run(new GlLernen(), args);
    }

    /** Der Logger der Klasse. */
    private Debug log;

    /**
     * Die PID des Messquerschnitts, für den als einziges gelernt wird. Wenn
     * {@code null}, dann wird für alle gelernt.
     */
    private String pidMQ;

    private String dbSpeicherPid;

    /** Die maximale Anzahl gleichzeitiger Threads beim Lernen. */
    private int maxThreads;

    private Archiv archiv;

    private BetriebsMeldungsSender betriebsmeldungsSender;

    private ObjektFactory objektFactory;

    private GanglinienSpeicherClient ganglinienSpeicher;

    private EreignisKalender ereignisKalender;

    /**
     * Gibt die maximale Anzahl gleichzeitiger Threads beim Lernen zurück.
     *
     * @return die maximale Threadanzahl.
     */
    public int getMaxThreads() {
        return maxThreads;
    }

    public GlLernen(ObjektFactory objektFactory, Archiv archiv, EreignisKalender ereignisKalender,
            GanglinienSpeicherClient ganglinienSpeicher, 
            GanglinienLernenParameterLeser ganglinienLernenParameterLeser,
            Collection<MessQuerschnitt> messquerschnitte) {

        this.objektFactory = objektFactory;
        this.archiv = archiv;
        this.ereignisKalender = ereignisKalender;
        this.ganglinienSpeicher = ganglinienSpeicher;
        new Verwaltung(this, ganglinienLernenParameterLeser, archiv, messquerschnitte);
    }

    /**
     * Allgemeine Initialisierung der Applikation.
     */
    private GlLernen() {
        // Standardkonstruktor beim Start als Standardapplikation
    }

    /**
     * Initialisiert die datenverteilerrelevanten Eigenschaften der Anwendung und
     * startet alle Submodule.
     */
    @Override
    public void initialize(final ClientDavInterface connection) throws Exception {
        MessageSender.getInstance()
                .setApplicationLabel(getClass().getSimpleName() + "_" + connection.getLocalApplicationObject().getId());

        DavChecker.checkConfiguration(connection);
        
        betriebsmeldungsSender = new DavBetriebsmeldungsSender(connection);
        objektFactory = new DavObjektFactory(connection);

        SystemObject speicherObjekt = ((DavObjektFactory) objektFactory).getGanglinienSpeicherObjekt(dbSpeicherPid);

        archiv = new DavArchiv(connection, betriebsmeldungsSender);

        ereignisKalender = new DavEreignisKalender((DavObjektFactory) objektFactory);
        ganglinienSpeicher = new DavGanglinienSpeicher((DavObjektFactory) objektFactory, speicherObjekt);

        final List<MessQuerschnitt> mqListe = new ArrayList<>();
        if (pidMQ == null) {
            for (final SystemObject so : connection.getDataModel().getType("typ.messQuerschnittAllgemein")
                    .getElements()) {
                mqListe.add(((DavObjektFactory) objektFactory).createMessQuerschnitt(so));
            }
        } else {
            SystemObject so = connection.getDataModel().getObject(pidMQ);
            mqListe.add(((DavObjektFactory) objektFactory).createMessQuerschnitt(so));
        }

        new Verwaltung(this, new DavGanglinienLernenParameterLeser(this, connection), archiv,
                mqListe);
        if (log != null) {
            log.info(() -> "Automatisches Lernen Ganglinien bereit.");
        }
    }

    /**
     * Neben den Kommandozeilenargumenten für den Datenverteiler werden folgende
     * unterstützt.
     * <ul>
     * <li>{@code -MQ=} hiermit kann zu Testzwecken das Lernen auf einen einzelnen
     * Messquerschnitt begrenzt werden (optional).</li>
     * <li>{@code -maxThreads=} hiermit die maximale Anzahl gleichzeitiger
     * Lernthreads vorgegeben werden (optional).</li>
     * </ul>
     */
    @Override
    public void parseArguments(final ArgumentList argumentList) throws Exception {
        Thread.setDefaultUncaughtExceptionHandler(this);
        log = Debug.getLogger();

        if (argumentList.hasArgument("-MQ")) {
            pidMQ = argumentList.fetchArgument("-MQ=").asNonEmptyString();
        } else {
            pidMQ = null;
        }

        if (argumentList.hasArgument("-maxThreads")) {
            maxThreads = Integer.parseInt(argumentList.fetchArgument("-maxThreads=").asNonEmptyString());
        } else {
            maxThreads = -1;
        }

        dbSpeicherPid = argumentList.fetchArgument("-speicherPid=").asString();

        GanglinienAspekte.initialisiere(argumentList);
    }

    /**
     * Gibt die Exception im Log und als Betriebsmeldung aus und beendet
     * anschließend die Applikation.
     */
    @Override
    public void uncaughtException(final Thread t, final Throwable e) {
        String message = MessageFormat.format(
                "Es ist ein unbekannter Fehler im Thread {0} eingetreten. Die Applikation SWE 5.1 Ganglinienprognose wird beendet. Stack Trace: {1}",
                t, getStackTrace(e));
        if (log == null) {
            System.err.println(message);
        } else {
            log.error(message);
            if (betriebsmeldungsSender != null) {
                betriebsmeldungsSender.fehler(null, message);
            }
        }
        System.exit(-1);
    }

    private String getStackTrace(final Throwable e) {
        final Writer result = new StringWriter();
        final PrintWriter printWriter = new PrintWriter(result);
        e.printStackTrace(printWriter);
        return result.toString();
    }

    public LocalDateTime getTime() {
        return objektFactory.getTime();
    }

    public GanglinienSpeicherClient getGanglinienSpeicher() {
        return ganglinienSpeicher;
    }

    public Archiv getArchiv() {
        return archiv;
    }

    public EreignisKalender getEreignisKalender() {
        return ereignisKalender;
    }

    public void fehler(MessQuerschnitt mq, String message) {
        if (betriebsmeldungsSender != null) {
            betriebsmeldungsSender.fehler(mq, message);
        }
    }

    public void fehler(String message) {
        fehler(null, message);
    }

    public void warnung(MessQuerschnitt mq, String message) {
        if (betriebsmeldungsSender != null) {
            betriebsmeldungsSender.fehler(mq, message);
        }
    }

    public void warnung(String message) {
        warnung(null, message);
    }

}
