/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.text.MessageFormat;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import de.bsvrz.iav.gllernen.gllernen.domain.GanglinienModellAutomatischesLernen;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.shared.CronJob;
import de.bsvrz.iav.gllib.gllib.shared.LocalDateTimeKonverter;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Führt einen Zyklus des Ganglinienlernens aus. Es wird für jeden
 * Messquerschnitt das Lernen mit den gleichen (aktuellen) Lernenparametern
 * durchgeführt.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public class LernTask extends CronJob {

    /** Der Logger der Klasse. */
    private final Debug log = Debug.getLogger();

    /** Der Lernparameter mit dem gelernt werden soll. */
    private GanglinienModellAutomatischesLernen lernParam;

    private final GlLernen applikation;

    private ExecutorService executorService;

    private final Verwaltung verwaltung;

    /**
     * Initialisiert den Lernprozess.
     *
     * @param applikation die Ganglinien-Lernen-Applikation
     * @param verwaltung  die Verwaltung der Lernaufgaben
     * @param lernParam   der Lernparameter.
     */
    public LernTask(GlLernen applikation, Verwaltung verwaltung, GanglinienModellAutomatischesLernen lernParam) {
        super(lernParam.getAktualisierungsintervall());

        this.applikation = applikation;
        this.verwaltung = verwaltung;
        this.lernParam = lernParam;

        executorService = Executors.newFixedThreadPool(verwaltung.getSlots());

        log.config(() -> MessageFormat.format("Es werden maximal {0} Messquerschnitte parallel gelernt.",
                verwaltung.getSemaphore().availablePermits()));
    }

    /**
     * Startet den Lernprozess für die angegebenen Messquerschnitte.
     */
    @Override
    public synchronized void run() {

        // Lernparameter zwischenspeichern, damit in diesem Zyklus für alle MQs
        // mit den selben Parameterdaten gelernt wird.
        GanglinienModellAutomatischesLernen param = getLernParam();

        // Lernzeitraum bestimmen
        LocalDateTime aktuelleZeit = applikation.getTime();
        LocalDateTime naechsterTagesStart = LocalDateTime.of(LocalDate.now(), LocalTime.MIDNIGHT).plusDays(1);
        LocalDateTime startLernzeitraum = verwaltung.getLetzteLernzeit();
        LocalDateTime endeLernzeitraum = naechsterTagesStart.minus(param.getDatenMindestalter());
        log.fine(() -> MessageFormat.format(
                "Initialisiere Lernen: aktuelleZeit={0}, startLernzeitraum={1}, endeLernzeitraum={2}", aktuelleZeit,
                startLernzeitraum, endeLernzeitraum));

        if (!endeLernzeitraum.isAfter(startLernzeitraum)) {
            log.fine(() -> "Lernen wird abgebrochen, da im Lernzeitraum bereits gelernt wurde.");
        } else {
            Intervall lernzeitraum = Intervall.of(LocalDateTimeKonverter.toMillis(startLernzeitraum),
                    LocalDateTimeKonverter.toMillis(endeLernzeitraum));

            log.info(
                    () -> MessageFormat.format("alle MQ: Starte neuen Lernzyklus für den Zeitraum {0}.", lernzeitraum));

            List<MessQuerschnitt> messQuerschnitte = new ArrayList<>(verwaltung.getMessquerschnitte());
            for (MessQuerschnitt messQuerschnitt : verwaltung.getMessquerschnitte()) {
                log.info(() -> MessageFormat.format("Starte das Lernen für MQ {0} ({1} von {2})", messQuerschnitt,
                        messQuerschnitte.indexOf(messQuerschnitt) + 1, verwaltung.getMessquerschnitte().size()));

                LernenMQ task = new LernenMQ(applikation, messQuerschnitt, lernzeitraum,
                        param.getMaxVergleichsAbstand());
                CompletableFuture.runAsync(task, executorService)
                        .whenComplete((result, exception) -> handleLernErgebnis(result, exception));
            }

            verwaltung.setLetzteLernzeit(endeLernzeitraum);
            log.info(() -> "alle " + messQuerschnitte.size() + " MQ: Das Lernen wurde erfolgreich beendet.");
        }
    }

    private void handleLernErgebnis(@SuppressWarnings("unused") Void result, Throwable exception) {
        if (exception != null) {
            log.warning("Fehler beim Lernen einer Ganglinie: " + exception.getLocalizedMessage());
        }
    }

    /**
     * Gibt den aktuellen Parameter des automatischen Ganglinienlernens zurück.
     *
     * @return der Lernparameter
     */
    public GanglinienModellAutomatischesLernen getLernParam() {
        return lernParam;
    }

    /**
     * Legt den Parameter des automatischen Ganglinienlernens fest.
     *
     * @param lernParam der neue Lernparameter.
     */
    public void setLernParam(final GanglinienModellAutomatischesLernen lernParam) {
        this.lernParam = lernParam;
        setPattern(lernParam.getAktualisierungsintervall());
    }
}
