/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2019 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.text.MessageFormat;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import de.bsvrz.iav.gllib.gllib.domain.EreignisTyp;
import de.bsvrz.iav.gllib.gllib.domain.EreignisTypLernParameter;
import de.bsvrz.iav.gllib.gllib.domain.GanglinieTyp;
import de.bsvrz.iav.gllib.gllib.domain.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.comm.GanglinienSpeicherClient;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlSchreibenAnfrage;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Lernt für den Zeitraum eines Tages eine Ganglinie.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Lernen {

    /** Der Logger der Klasse. */
    private final Debug log = Debug.getLogger();

    /** Die Ausgangsdaten des Lernens. */
    private final Ausgangsdaten ausgangsdaten;

    /** Der maximale Vergleichsabstand für das Verschmelzen von Ganglinien. */
    private final int maxVergleichsAbstand;

    /** Der Ereignistyp, für den eine Ganglinien gelernt wird. */
    private EreignisTyp ereignisTyp;

    /**
     * Der Index der historische Ganglinie, die zum Lernen/Verschmelzen in Frage
     * kommt. Der Wert -1 signalisiert, dass keine passende historische Ganglinie
     * gefunden werden konnte.
     *
     * FIXME Sollte das nicht ein "int" sein, wenn im Fehlerfall -1 geliefert wird?
     */
    private Integer historischeGanglinieIndex = -1;

    /**
     * Der Abstand der gefundenen historischen Ganglinie zur Analyseganglinie.
     */
    private double historischeGanglinieAbstand = Double.MAX_VALUE;

    private final GlLernen applikation;

    /**
     * Initialisierung den Lernprozess.
     *
     * @param applikation          die Ganglinien-Lernen-Applikation
     * @param mq                   der Messquerschnitt für den die Ganglinien
     *                             gelernt werden.
     * @param lernzeitraum         der Zeitraum in dem gelernt werden soll.
     * @param maxVergleichsAbstand der maximale Vergleichsabstand für das
     *                             Verschmelzen von Ganglinien.
     */
    public Lernen(GlLernen applikation, final MessQuerschnitt mq, final Intervall lernzeitraum,
            final int maxVergleichsAbstand) {
        this.applikation = applikation;
        this.maxVergleichsAbstand = maxVergleichsAbstand;
        ausgangsdaten = new Ausgangsdaten(applikation, mq, lernzeitraum);
    }

    /**
     * Führt das Lernen aus.
     */
    public void lerne() {
        log.info(() -> MessageFormat.format("{0}: Starte neuen Lernzyklus für den Zeitraum {1}.", ausgangsdaten.getMq(),
                ausgangsdaten.getLernzeitraum()));

        if (!ausgangsdaten.abrufen()) {
            log.info(() -> MessageFormat.format("{0}: Das Lernen wird abgebrochen.", ausgangsdaten.getMq()));
            return;
        }

        // Ereignistyp zum Lernen bestimmen
        ereignisTyp = bestimmeLernEreignisTyp(false);
        if (ereignisTyp == null) {
            log.info(() -> MessageFormat.format("{0}: Das Lernen wird abgebrochen.", ausgangsdaten.getMq()));
            return;
        }

        // Nun kann gelernt werden
        MqGanglinie gelernt = null;
        GanglinieTyp ganglinieTyp = ereignisTyp.getLernParameter().getGanglinieTyp();
        if (ganglinieTyp == GanglinieTyp.ADDITIV || ganglinieTyp == GanglinieTyp.MULTIPLIKATIV) {
            gelernt = lerneRelativeGanglinie();
        }
        if (ganglinieTyp == GanglinieTyp.ABSOLUT || gelernt == null) {
            if (ganglinieTyp != GanglinieTyp.ABSOLUT) {
                ereignisTyp = bestimmeLernEreignisTyp(true);
                if (ereignisTyp == null) {
                    log.info(() -> MessageFormat.format("{0}: Das Lernen wird abgebrochen.", ausgangsdaten.getMq()));
                    return;
                }
            }
            gelernt = lerneAbsoluteGanglinie();
        }

        if (gelernt != null && gelernt.hatMesswerte()) {
            sichereGanglinie(gelernt);
        }
    }

    /**
     * Sucht den Ereignistyp mit der höchsten Priorität. Der ausgewählte Ereignistyp
     * besitzt optional eine Ausschlussliste. Liegen Ereignistypen aus dieser
     * Ausschlussliste im Lernzyklus an, wird der Ereignistyp mit der nächst
     * niedrigeren Priorität ausgewählt.
     *
     * @param nurAbsolute wenn {@code true}, dann werden nur Ereignistypen
     *                    berücksichtigt, zu denen eine absolute Ganglinie gelernt
     *                    werden soll. Wenn {@code false}, dann wird nicht zwischen
     *                    absolut und relativ unteschieden.
     * @return der Ereignistyp oder {@code null}, wenn keiner gefunden werden kann.
     */
    private EreignisTyp bestimmeLernEreignisTyp(final boolean nurAbsolute) {

        Queue<EreignisTyp> liste = new LinkedList<>(ausgangsdaten.getEreignisTypen());
        EreignisTyp typ = null;

        while ((typ == null) && (liste.size() > 0)) {

            // Ereignistyp mit der höchsten Priorität bestimmen
            typ = liste.poll();

            if (nurAbsolute && (typ.getLernParameter().getGanglinieTyp() != GanglinieTyp.ABSOLUT)) {
                // Keine Typ für absolute Ganglinie, ergo weitersuchen
                typ = null;
                continue;
            }

            // Prüfen ob Ereignistypen anliegen, die in Ausschlussliste stehen
            for (final EreignisTyp et : typ.getLernParameter().getAusschlussliste()) {
                if (ausgangsdaten.getEreignisTypen().contains(et)) {
                    // Typ kann nicht verwendet werden, ergo weitersuchen
                    typ = null;
                    break;
                }
            }
        }

        if (typ != null) {
            log.fine(MessageFormat.format("{0}: Der Ereignistyp mit der höchsten Priorität ist {1}.",
                    ausgangsdaten.getMq(), typ));
        } else {
            String message = MessageFormat.format("{0}: Es konnte kein Ereignistyp zum Lernen ausgewählt werden.",
                    ausgangsdaten.getMq());
            log.warning(() -> message);
            applikation.warnung(ausgangsdaten.getMq(), message);
        }
        return typ;
    }

    /**
     * Lernt eine absolute Ganglinie.
     */
    private MqGanglinie lerneAbsoluteGanglinie() {

        log.fine(() -> MessageFormat.format("{0}: Lerne absolute Ganglinie ...", ausgangsdaten.getMq()));

        bestimmeAbsoluteGanglinieMitGeringstemAbstand(ereignisTyp);

        EreignisTypLernParameter lernParameter = ereignisTyp.getLernParameter();

        if (lernParameter.getMaximalerAbstand() < maxVergleichsAbstand) {
            String message = MessageFormat.format(
                    "{0}: Der Trigger für das Anlegen neuer Ganglinien für den Ereignistyp {1} ist kleiner als der globale Trigger für Warningen bei zu großen Abstand. Es wird nie eine Warning wegen zu großem Abstand für diesen Ereignistyp geben.",
                    ausgangsdaten.getMq(), ereignisTyp);
            log.warning(() -> message);
            applikation.warnung(ausgangsdaten.getMq(), message);
        }

        MqGanglinie ganglinie = null;
        MqGanglinie archivGanglinie = ausgangsdaten.getArchivGanglinie(lernParameter.getDarstellungsVerfahren());
        if (!archivGanglinie.hatMesswerte()) {
            String message = MessageFormat.format("{0}: Im Lernzeitraumen liegen keine Archivdaten vor.",
                    ausgangsdaten.getMq());
            log.warning(() -> message);
            applikation.warnung(ausgangsdaten.getMq(), message);
            return archivGanglinie;
        }

        if (historischeGanglinieAbstand <= lernParameter.getMaximalerAbstand()) {
            // Es wurde eine passende Ganglinien zum Verschmelzen gefunden und
            // der Abstand ist nicht zu groß
            long gewicht;
            MqGanglinie historischeGl;

            historischeGl = ausgangsdaten.getHistorischeGanglinien().get(historischeGanglinieIndex);

            // Gewicht für das Verschmelzen bestimmen
            gewicht = historischeGl.getAnzahlVerschmelzungen();
            if (gewicht > lernParameter.getMaxWichtungsFaktor()) {
                gewicht = lernParameter.getMaxWichtungsFaktor();
            }
            if (gewicht <= 0) {
                gewicht = 1;
            }

            long logGewicht = gewicht;
            log.fine(() -> MessageFormat.format(
                    "{0}: Die historische Ganglinie wird mit dem Gewicht {1} mit der Analyseganglinie verschmolzen.",
                    ausgangsdaten.getMq(), logGewicht));

            // Ganglinie veschmelzen
            ganglinie = GanglinienMQOperationen.verschmelze(applikation.getTime(), archivGanglinie, historischeGl,
                    gewicht);
            ausgangsdaten.getHistorischeGanglinien().set(historischeGanglinieIndex, ganglinie);
        } else {
            // Neue absolute Ganglinie anlegen
            if (historischeGanglinieIndex >= 0) {
                log.fine(() -> MessageFormat.format(
                        "{0}: Der maximale Abstand zur historischen Ganglinie wird überschritten. Es wird eine neue Ganglinie angelegt.",
                        ausgangsdaten.getMq()));
            } else {
                log.fine(() -> MessageFormat.format("{0}: Es wird eine neue Ganglinie angelegt.",
                        ausgangsdaten.getMq()));
            }

            ganglinie = archivGanglinie.erzeugeKopieMitEreignisTyp(ereignisTyp);
            ausgangsdaten.getHistorischeGanglinien().add(ganglinie);
            if ((historischeGanglinieAbstand != Double.MAX_VALUE)
                    && (historischeGanglinieAbstand > maxVergleichsAbstand)) {
                String message = MessageFormat.format(
                        "{0}: Der Vergleichsabstand {1} der neuen Ganglinie überschreitet das erlaubte Maximum von {2}.",
                        ausgangsdaten.getMq(), historischeGanglinieAbstand, maxVergleichsAbstand);
                log.warning(() -> message);
                applikation.warnung(ausgangsdaten.getMq(), message);
            }
        }

        return ganglinie;
    }

    /**
     * Sucht die absolute Ganglinie mit dem geringsten komplexen Abstand zur
     * Analyseganglinie. Es wird der Index der Ganglinie und deren Abstand zur
     * Analyseganglinie bestimmt.
     *
     * @param typ der Ereignistyp, zu dem eine absolute Ganglinie gesucht wird.
     * @see #historischeGanglinieIndex
     * @see #historischeGanglinieAbstand
     */
    private void bestimmeAbsoluteGanglinieMitGeringstemAbstand(final EreignisTyp typ) {
        int index;

        log.finer(() -> MessageFormat.format(
                "{0}: Suche die Ganglinie mit dem Ereignistyp {1} und dem geringsten komplexen Abstand zur Analyseganglinie.",
                ausgangsdaten.getMq(), typ));

        EreignisTypLernParameter paramTypLernen = typ.getLernParameter();
        assert paramTypLernen
                .getGanglinieTyp() == GanglinieTyp.ABSOLUT : "Für ein Tagesereignis darf nur eine absolute Ganglinie gelernt werden.";

        // Index der historischen Ganglinie mit dem geringsten Abstand bestimmen
        double abstand = Double.MAX_VALUE;
        index = -1;

        MqGanglinie archivGanglinie = ausgangsdaten.getArchivGanglinie(paramTypLernen.getDarstellungsVerfahren());
        if (archivGanglinie.hatMesswerte()) {
            for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien().size(); ++i) {

                MqGanglinie g = ausgangsdaten.getHistorischeGanglinien().get(i);
                if (g.getEreignisTyp().equals(typ)) {
                    double d = GanglinienMQOperationen.komplexerAbstand(archivGanglinie, g,
                            paramTypLernen.getVergleichsSchrittweite().toMillis());
                    int logIndex = i;
                    log.finer(() -> MessageFormat.format(
                            "{0}: Komplexer Abstand der untersuchten Ganglinie: index={1}, abstand={2}",
                            ausgangsdaten.getMq(), logIndex, d));
                    if (d < abstand) {
                        abstand = d;
                        index = i;
                    }
                }
            }
        }

        if (index >= 0) {
            int logIndex = index;
            double logAbstand = abstand;
            log.fine(() -> MessageFormat.format(
                    "{0}: Komplexer Abstand der gefundenen Ganglinie: index={1}, abstand={2}, {3}",
                    ausgangsdaten.getMq(), logIndex, logAbstand,
                    GanglinienMQOperationen.formatierterText(ausgangsdaten.getHistorischeGanglinien().get(logIndex))));
        } else {
            log.fine(() -> MessageFormat.format("{0}: Es existiert noch keine historische Ganglinie.",
                    ausgangsdaten.getMq()));
        }

        historischeGanglinieIndex = index;
        historischeGanglinieAbstand = abstand;
    }

    /**
     * Sichert die Ganglinienliste im Datenverteiler.
     * 
     * @param gelernt
     */
    private void sichereGanglinie(MqGanglinie ganglinie) {

        // Eigenschaften ergänzen bzw. aktualisieren
        assert ganglinie != null : "Irgendeine Ganglinie muss erzeugt oder verschmolzen werden.";
        assert ereignisTyp != null : "Der Ereignistyp muss ausgewählt sein.";

        EreignisTypLernParameter param = ereignisTyp.getLernParameter();

        ganglinie.setEreignisTyp(ereignisTyp);
        ganglinie.setApproximationDaK(param.getDarstellungsVerfahren());

        log.finer(() -> MessageFormat.format("{0}: Gelernte Ganglinie: {1}", ausgangsdaten.getMq(),
                GanglinienMQOperationen.formatierterText(ganglinie)));
        log.finest(() -> MessageFormat.format("{0}: Neue Liste der historischen Ganglinien: {1}", ausgangsdaten.getMq(),
                ausgangsdaten.getHistorischeGanglinien()));

        while (zaehleGanglinien(ausgangsdaten.getHistorischeGanglinien(), ereignisTyp) > param.getMaxGanglinien()) {
            // Die älteste Ganglinie für den Ereignistyp löschen
            LocalDateTime t = LocalDateTime.MAX;
            int index = -1;

            for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien().size(); i++) {
                final MqGanglinie mqGanglinie = ausgangsdaten.getHistorischeGanglinien().get(i);
                if (ereignisTyp.equals(mqGanglinie.getEreignisTyp())) {
                    if (t.isAfter(mqGanglinie.getLetzteVerschmelzung())) {
                        t = mqGanglinie.getLetzteVerschmelzung();
                        index = i;
                    }
                }
            }
            if (index >= 0) {
                int logIndex = index;
                log.fine(() -> MessageFormat.format(
                        "{0}: Anzahl maximaler Ganglinie für Ereignistyp {1} überschritten. Die älteste Ganglinie wird vergessen: {2}",
                        ausgangsdaten.getMq(), ereignisTyp, GanglinienMQOperationen
                                .formatierterText(ausgangsdaten.getHistorischeGanglinien().get(logIndex))));
                ausgangsdaten.getHistorischeGanglinien().remove(index);
            }
        }

        // Ganglinienliste des MQ aktualisieren
        GanglinienSpeicherClient ganglinienSpeicher = applikation.getGanglinienSpeicher();
        GlSchreibenAnfrage anfrage = GlSchreibenAnfrage.of(ganglinienSpeicher.getAnfrager(),
                ganglinienSpeicher.getAnfrageId(), ausgangsdaten.getHistorischeGanglinien());
		ganglinienSpeicher.schreibeGanglinien(anfrage, a -> {
            log.finer(() -> "Ganglinien für MQ '" + ausgangsdaten.getMq() + "' geschrieben");
        }, f -> {
            log.warning(() -> f.getFehlerMeldungen().toString());
            applikation.warnung(ausgangsdaten.getMq(), f.getFehlerMeldungen().toString());
        });
    }

    /**
     * Zählt die Ganglinien eines Ereignistyps in einer Ganglinienliste.
     *
     * @param ganglinien  eine Ganglinienliste.
     * @param ereignisTyp ein Ereignistyp.
     * @return die Anzahl der Ganglinien zum Ereignistyp.
     */
    private static int zaehleGanglinien(final List<MqGanglinie> ganglinien, final EreignisTyp ereignisTyp) {
        int i;

        i = 0;
        for (final MqGanglinie g : ganglinien) {
            if (ereignisTyp.equals(g.getEreignisTyp())) {
                ++i;
            }
        }
        return i;
    }

    /**
     * Lernt eine relative Ganglinie (additiv oder multiplikativ).
     *
     * @return {@code true}, wenn eine Ganglinie gelernt wurde.
     */
    private MqGanglinie lerneRelativeGanglinie() {
        final EreignisTyp typAbsolut;
        final List<MqGanglinie> listeAbsoluteGl, listeRelativeGl;
        final Map<Integer, Integer> matchingHilfe;
        final Map<Integer, Integer> historischeGlHilfe;
        final List<MqGanglinie> matching;
        final GanglinienMQOperationen.MqGangliniePatternMatchingErgebnis matchingErg;

        log.info(() -> MessageFormat.format("{0}: Lerne relative Ganglinie ...", ausgangsdaten.getMq()));

        typAbsolut = bestimmeBezugsereignis(ereignisTyp);
        if (typAbsolut == null) {
            log.finer(() -> MessageFormat.format(
                    "{0}: Es gibt keinen Ereignistyp, auf den sich die zu lernende relative Ganglinie beziehen kann.",
                    ausgangsdaten.getMq()));
            return null;
        }

        final EreignisTypLernParameter param = ereignisTyp.getLernParameter();
        MqGanglinie archivGanglinie = ausgangsdaten.getArchivGanglinie(param.getDarstellungsVerfahren());
        if (!archivGanglinie.hatMesswerte()) {
            String message = MessageFormat.format("{0}: Im Lernzeitraumen liegen keine Archivdaten vor.",
                    ausgangsdaten.getMq());
            log.warning(() -> message);
            applikation.warnung(ausgangsdaten.getMq(), message);
            return archivGanglinie;
        }

        EreignisTypLernParameter paramAbsolut = typAbsolut.getLernParameter();
        log.finer(() -> MessageFormat.format(
                "{0}: Die zu lernende relative Ganglinie bezieht sich auf eine absolute Ganglinie für den Ereignistyp {1}.",
                ausgangsdaten.getMq(), typAbsolut));

        // Bestimme alle historischen Ganglinien zum Bezugsereignis
        listeAbsoluteGl = new ArrayList<>();
        for (final MqGanglinie g : ausgangsdaten.getHistorischeGanglinien()) {
            if (g.getEreignisTyp().equals(typAbsolut)) {
                listeAbsoluteGl.add(g);
            }
        }
        assert listeAbsoluteGl.size() > 0 : "Es müssen absolute Ganglinien existieren, da ein Bezugstyp existiert.";

        // Ganglinien-Index -> Index in historischen Ganglinien
        historischeGlHilfe = new HashMap<>();

        // Bestimme alle historischen Ganglinien zum zu lernenden Ereignis
        listeRelativeGl = new ArrayList<>();
        for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien().size(); ++i) {
            final MqGanglinie g;

            g = ausgangsdaten.getHistorischeGanglinien().get(i);
            if (g.getEreignisTyp().equals(ereignisTyp)) {
                listeRelativeGl.add(g);
                historischeGlHilfe.put(listeRelativeGl.size() - 1, i);
            }
        }

        // Matching-Index -> Ganglinien-Index
        matchingHilfe = new HashMap<>();

        // Pattern-Matching ausführen
        matching = new ArrayList<>();
        for (int i = 0; i < listeAbsoluteGl.size(); ++i) {
            MqGanglinie absoluteGl;

            absoluteGl = listeAbsoluteGl.get(i);
            if (listeRelativeGl.isEmpty()) {
                matching.add(absoluteGl.erzeugeKopie());
                matchingHilfe.put(matching.size() - 1, -1);
            } else {
                for (int j = 0; j < listeRelativeGl.size(); ++j) {
                    final MqGanglinie relativeGl;
                    MqGanglinie g;

                    relativeGl = listeRelativeGl.get(j);
                    if (relativeGl.getTyp() == GanglinieTyp.ADDITIV) {
                        g = GanglinienMQOperationen.addiere(absoluteGl, relativeGl);
                    } else if (relativeGl.getTyp() == GanglinieTyp.MULTIPLIKATIV) {
                        g = GanglinienMQOperationen.multipliziere(absoluteGl, relativeGl);
                    } else {
                        assert false : "Die Ganglinie zu " + ereignisTyp + " muss relativ sein, war aber "
                                + relativeGl.getTyp();
                        continue;
                    }
                    matching.add(g);
                    matchingHilfe.put(matching.size() - 1, j);
                }
            }
        }
        matchingErg = GanglinienMQOperationen.patternMatching(archivGanglinie, matching,
                paramAbsolut.getMatchingIntervallVor().toMillis(), paramAbsolut.getMatchingIntervallNach().toMillis(),
                paramAbsolut.getMatchingSchrittweite().toMillis());

        log.fine(() -> MessageFormat.format("{0}: Das Ergebnis des Pattern-Matching: {1}", ausgangsdaten.getMq(),
                matchingErg));

        historischeGanglinieIndex = historischeGlHilfe.get(matchingHilfe.get(matchingErg.getIndex()));

        log.fine(() -> MessageFormat.format(
                "{0}: Es wird die historische Ganglinie mit dem Index {1} aktualisiert: {2}", ausgangsdaten.getMq(),
                historischeGanglinieIndex, ausgangsdaten.getHistorischeGanglinien().get(historischeGanglinieIndex)));

        // Die gelernte Ganglinie bestimmen

        MqGanglinie ganglinie = null;

        archivGanglinie.setApproximationDaK(
                ausgangsdaten.getHistorischeGanglinien().get(historischeGanglinieIndex).getApproximationDaK());
        switch (ausgangsdaten.getHistorischeGanglinien().get(historischeGanglinieIndex).getTyp()) {
        case ADDITIV:
            ganglinie = GanglinienMQOperationen.subtrahiere(archivGanglinie, matchingErg.getGanglinie());
            break;
        case MULTIPLIKATIV:
            ganglinie = GanglinienMQOperationen.dividiere(archivGanglinie, matchingErg.getGanglinie());
            break;
        default:
            assert false : "Der Typ der zu lernenden Ganglinie muss relativ sein.";
        }

        MqGanglinie logGanglinie = ganglinie;
        log.fine(() -> MessageFormat.format("{0}: Die zu lernende Ganglinie ist: {1}", ausgangsdaten.getMq(),
                logGanglinie));

        // Neue Ganglinie oder mit vorhandener verschmelzen?
        if (matchingErg.getAbstand() < param.getMaximalerMatchingFehler()) {
            // Es wurde eine passende Ganglinien zum Verschmelzen gefunden und
            // der Abstand ist nicht zu groß
            int gewicht;
            MqGanglinie historischeGl;

            historischeGl = ausgangsdaten.getHistorischeGanglinien().get(historischeGanglinieIndex);

            // Gewicht für das Verschmelzen bestimmen
            gewicht = (int) historischeGl.getAnzahlVerschmelzungen();
            if (gewicht > param.getMaxWichtungsFaktor()) {
                gewicht = param.getMaxWichtungsFaktor();
            }
            if (gewicht == 0) {
                gewicht = 1;
            }

            int logGewicht = gewicht;
            log.fine(() -> MessageFormat.format(
                    "{0}: Die historische Ganglinie wird mit dem Gewicht {1} mit der Analyseganglinie verschmolzen.",
                    ausgangsdaten.getMq(), logGewicht));

            // Ganglinie veschmelzen
            ganglinie = GanglinienMQOperationen.verschmelze(applikation.getTime(), ganglinie, historischeGl, gewicht);
            ausgangsdaten.getHistorischeGanglinien().set(historischeGanglinieIndex, ganglinie);
        } else {
            // Neue relative Ganglinie anlegen
            if (historischeGanglinieIndex != null) {
                log.fine(() -> MessageFormat.format(
                        "{0}: Der maximale Abstand zur historischen Ganglinie wird überschritten. Es wird eine neue Ganglinie angelegt.",
                        ausgangsdaten.getMq()));
            } else {
                log.fine(() -> MessageFormat.format("{0}: Es wird eine neue Ganglinie angelegt.",
                        ausgangsdaten.getMq()));
            }

            ausgangsdaten.getHistorischeGanglinien().add(ganglinie);
            if ((historischeGanglinieAbstand != Double.MAX_VALUE)
                    && (historischeGanglinieAbstand > maxVergleichsAbstand)) {

                String message = MessageFormat.format(
                        "{0}: Der Vergleichsabstand {1} der neuen Ganglinie überschreitet das erlaubte Maximum von {2}.",
                        ausgangsdaten.getMq(), historischeGanglinieAbstand, maxVergleichsAbstand);
                log.warning(() -> message);
                applikation.warnung(ausgangsdaten.getMq(), message);
            }
        }

        return ganglinie;
    }

    /**
     * Bestimmt den Ereignistyp mit einer absoluten Ganglinie im Lernzyklus mit der
     * höchsten Priorität auf den sich ein Ereignis einer relativen Ganglinie
     * beziehen kann. Der ausgewählte Ereignistyp besitzt optional eine
     * Ausschlussliste. Liegen Ereignistypen aus dieser Ausschlussliste im
     * Lernzyklus an, wird der Ereignistyp mit der nächst niedrigeren Priorität
     * ausgewählt.
     *
     * @param relTyp der relative Ereignistyp zu dem ein passender absoluter
     *               Ereignistyp gesucht ist.
     * @return der gefundene Ereignistyp oder {@code null}, wenn keiner bestimmt
     *         werden konnte.
     */
    private EreignisTyp bestimmeBezugsereignis(final EreignisTyp relTyp) {

        EreignisTyp typ;
        LinkedList<EreignisTyp> liste;

        liste = new LinkedList<>(ausgangsdaten.getEreignisTypen());

        EreignisTypLernParameter relTypParam = relTyp.getLernParameter();

        typ = null;
        while (liste.size() > 0) {

            // Ereignistyp mit der höchsten Priorität bestimmen
            typ = liste.poll();
            EreignisTypLernParameter param = typ.getLernParameter();

            // Tagesereignis gefunden?
            if (param.getGanglinieTyp() != GanglinieTyp.ABSOLUT) {
                typ = null;
                continue;
            }

            // Prüfen ob Ereignistypen anliegen, die in Ausschlussliste stehen.
            for (final EreignisTyp et : param.getAusschlussliste()) {
                if (ausgangsdaten.getEreignisTypen().contains(et)) {
                    typ = null;
                    continue;
                }
            }

            // Ist das Tagesereignis mit dem relativen Ereignis kombinierbar?
            final Collection<EreignisTyp> bezugsereignistypen = relTypParam.getBezugsTypen();
            if (!((bezugsereignistypen.isEmpty()) || (bezugsereignistypen.contains(typ)))) {
                typ = null;
                continue;
            }

            // Gibt es eine Ganglinie für das Tagesereignis?
            boolean gefunden = false;
            for (final MqGanglinie g : ausgangsdaten.getHistorischeGanglinien()) {
                if (g.getEreignisTyp().equals(typ)) {
                    gefunden = true;
                    break;
                }
            }
            if (gefunden) {
                break;
            }
            typ = null;
        }

        return typ;
    }

}
